@@ -1,7 +1,7 @@
 # vim: ts=8:sts=8:sw=8:noexpandtab
 #
 # This file is part of ReText
-# Copyright: 2015 Dmitry Shachnev
+# Copyright: 2015-2017 Dmitry Shachnev
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -16,17 +16,14 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+from os.path import exists, splitext
 from markups import get_markup_for_file_name, find_markup_class_by_name
 from markups.common import MODULE_HOME_PAGE
 
 from ReText import app_version, globalSettings, converterprocess
 from ReText.editor import ReTextEdit
 from ReText.highlighter import ReTextHighlighter
-
-try:
-	from ReText.fakevimeditor import ReTextFakeVimHandler
-except ImportError:
-	ReTextFakeVimHandler = None
+from ReText.preview import ReTextPreview
 
 try:
 	import enchant
@@ -34,13 +31,18 @@ except ImportError:
 	enchant = None
 
 from PyQt5.QtCore import pyqtSignal, Qt, QDir, QFile, QFileInfo, QPoint, QTextStream, QTimer, QUrl
-from PyQt5.QtGui import QDesktopServices, QTextCursor, QTextDocument
-from PyQt5.QtWidgets import QTextBrowser, QTextEdit, QSplitter
+from PyQt5.QtGui import QTextCursor, QTextDocument
+from PyQt5.QtWidgets import QTextEdit, QSplitter
 
 try:
-	from ReText.webkitpreview import ReTextWebPreview
+	from ReText.webkitpreview import ReTextWebKitPreview
 except ImportError:
-	ReTextWebPreview = None
+	ReTextWebKitPreview = None
+
+try:
+	from ReText.webenginepreview import ReTextWebEnginePreview
+except ImportError:
+	ReTextWebEnginePreview = None
 
 PreviewDisabled, PreviewLive, PreviewNormal = range(3)
 
@@ -84,7 +86,7 @@ class ReTextTab(QSplitter):
 		self.editBox.textChanged.connect(self.triggerPreviewUpdate)
 		self.editBox.undoAvailable.connect(parent.actionUndo.setEnabled)
 		self.editBox.redoAvailable.connect(parent.actionRedo.setEnabled)
-		self.editBox.copyAvailable.connect(parent.actionCopy.setEnabled)
+		self.editBox.copyAvailable.connect(parent.enableCopy)
 
 		# Give both boxes a minimum size so the minimumSizeHint will be
 		# ignored when splitter.setSizes is called below
@@ -103,7 +105,7 @@ class ReTextTab(QSplitter):
 
 	def createPreviewBox(self, editBox):
 
-		# Use closures to avoid a hard reference from ReTextWebPreview
+		# Use closures to avoid a hard reference from ReTextWebKitPreview
 		# to self, which would keep the tab and its resources alive
 		# even after other references to it have disappeared.
 
@@ -118,10 +120,14 @@ class ReTextTab(QSplitter):
 			rect = doc.documentLayout().blockBoundingRect(block)
 			return rect.top()
 
-		if globalSettings.useWebKit:
-			preview = ReTextWebPreview(editBox,
-			                           editorPositionToSourceLine,
-			                           sourceLineToEditorPosition)
+		if ReTextWebKitPreview and globalSettings.useWebKit:
+			preview = ReTextWebKitPreview(self,
+			                              editorPositionToSourceLine,
+			                              sourceLineToEditorPosition)
+		elif ReTextWebEnginePreview and globalSettings.useWebEngine:
+			preview = ReTextWebEnginePreview(self,
+			                                 editorPositionToSourceLine,
+			                                 sourceLineToEditorPosition)
 		else:
 			preview = ReTextPreview(self)
 
@@ -275,37 +281,95 @@ class ReTextTab(QSplitter):
 		self.editBox.setVisible(self.previewState < PreviewNormal)
 		self.previewBox.setVisible(self.previewState > PreviewDisabled)
 
+	def rebuildPreviewBox(self):
+		self.previewBox.disconnectExternalSignals()
+		self.previewBox.setParent(None)
+		self.previewBox.deleteLater()
+		self.previewBox = self.createPreviewBox(self.editBox)
+		self.previewBox.setMinimumWidth(125)
+		self.addWidget(self.previewBox)
+		self.setSizes((50, 50))
+		self.triggerPreviewUpdate()
+		self.updateBoxesVisibility()
+
+	def detectFileEncoding(self, fileName):
+		'''
+		Detect content encoding of specific file.
+
+		It will return None if it can't determine the encoding.
+		'''
+		try:
+			import chardet
+		except ImportError:
+			return
+
+		with open(fileName, 'rb') as inputFile:
+			raw = inputFile.read(2048)
+
+		result = chardet.detect(raw)
+		if result['confidence'] > 0.9:
+			if result['encoding'].lower() == 'ascii':
+				# UTF-8 files can be falsely detected as ASCII files if they
+				# don't contain non-ASCII characters in first 2048 bytes.
+				# We map ASCII to UTF-8 to avoid such situations.
+				return 'utf-8'
+			return result['encoding']
+
 	def readTextFromFile(self, fileName=None, encoding=None):
 		previousFileName = self._fileName
 		if fileName:
 			self._fileName = fileName
+
+		# Only try to detect encoding if it is not specified
+		if encoding is None and globalSettings.detectEncoding:
+			encoding = self.detectFileEncoding(self._fileName)
+
+		# TODO: why do we open the file twice: for detecting encoding
+		# and for actual read? Can we open it just once?
 		openfile = QFile(self._fileName)
 		openfile.open(QFile.ReadOnly)
 		stream = QTextStream(openfile)
 		encoding = encoding or globalSettings.defaultCodec
 		if encoding:
 			stream.setCodec(encoding)
+			# If encoding is specified or detected, we should save the file with
+			# the same encoding
+			self.editBox.document().setProperty("encoding", encoding)
+
 		text = stream.readAll()
 		openfile.close()
 
-		modified = bool(encoding) and (self.editBox.toPlainText() != text)
-		self.editBox.setPlainText(text)
-		self.editBox.document().setModified(modified)
-
 		if previousFileName != self._fileName:
 			self.updateActiveMarkupClass()
+
+		self.editBox.setPlainText(text)
+		self.editBox.document().setModified(False)
+
+		if previousFileName != self._fileName:
 			self.fileNameChanged.emit()
 
 	def writeTextToFile(self, fileName=None):
 		# Just writes the text to file, without any changes to tab object
 		# Used directly for i.e. export extensions
+
+		# Get text from the cursor to avoid tweaking special characters,
+		# see https://bugreports.qt.io/browse/QTBUG-57552 and
+		# https://github.com/retext-project/retext/issues/216
+		cursor = self.editBox.textCursor()
+		cursor.select(QTextCursor.Document)
+		text = cursor.selectedText().replace('\u2029', '\n')
+
 		savefile = QFile(fileName or self._fileName)
 		result = savefile.open(QFile.WriteOnly)
 		if result:
 			savestream = QTextStream(savefile)
-			if globalSettings.defaultCodec:
-				savestream.setCodec(globalSettings.defaultCodec)
-			savestream << self.editBox.toPlainText()
+
+			# Save the file with original encoding
+			encoding = self.editBox.document().property("encoding")
+			if encoding is not None:
+				savestream.setCodec(encoding)
+
+			savestream << text
 			savefile.close()
 		return result
 
@@ -324,56 +388,57 @@ class ReTextTab(QSplitter):
 
 		return result
 
-	def installFakeVimHandler(self):
-		if ReTextFakeVimHandler:
-			fakeVimEditor = ReTextFakeVimHandler(self.editBox, self)
-			fakeVimEditor.setSaveAction(self.actionSave)
-			fakeVimEditor.setQuitAction(self.actionQuit)
-			# TODO: action is bool, really call remove?
-			self.p.actionFakeVimMode.triggered.connect(fakeVimEditor.remove)
-
-	def find(self, text, flags):
+	def find(self, text, flags, replaceText=None, wrap=False):
 		cursor = self.editBox.textCursor()
-		newCursor = self.editBox.document().find(text, cursor, flags)
-		if not newCursor.isNull():
-			self.editBox.setTextCursor(newCursor)
-			return True
-		cursor.movePosition(QTextCursor.End if (flags & QTextDocument.FindBackward) else QTextCursor.Start)
-		newCursor = self.editBox.document().find(text, cursor, flags)
+		if wrap and flags & QTextDocument.FindBackward:
+			cursor.movePosition(QTextCursor.End)
+		elif wrap:
+			cursor.movePosition(QTextCursor.Start)
+		if replaceText is not None and cursor.selectedText() == text:
+			newCursor = cursor
+		else:
+			newCursor = self.editBox.document().find(text, cursor, flags)
 		if not newCursor.isNull():
+			if replaceText is not None:
+				newCursor.insertText(replaceText)
+				newCursor.movePosition(QTextCursor.Left, QTextCursor.MoveAnchor, len(replaceText))
+				newCursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, len(replaceText))
 			self.editBox.setTextCursor(newCursor)
 			return True
+		if not wrap:
+			return self.find(text, flags, replaceText, True)
 		return False
 
+	def replaceAll(self, text, replaceText):
+		cursor = self.editBox.textCursor()
+		cursor.beginEditBlock()
+		cursor.movePosition(QTextCursor.Start)
+		flags = QTextDocument.FindFlags()
+		cursor = lastCursor = self.editBox.document().find(text, cursor, flags)
+		while not cursor.isNull():
+			cursor.insertText(replaceText)
+			lastCursor = cursor
+			cursor = self.editBox.document().find(text, cursor, flags)
+		if not lastCursor.isNull():
+			lastCursor.movePosition(QTextCursor.Left, QTextCursor.MoveAnchor, len(replaceText))
+			lastCursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, len(replaceText))
+			self.editBox.setTextCursor(lastCursor)
+		self.editBox.textCursor().endEditBlock()
+		return not lastCursor.isNull()
 
-class ReTextPreview(QTextBrowser):
-	"""
-	When links like [test](test) are clicked, the file test.md is opened.
-	It has to be located next to the current opened file.
-	Relative pathes like [test](../test) or [test](folder/test) are also possible.
-	"""
+	def openSourceFile(self, fileToOpen):
+		"""Finds and opens the source file for link target fileToOpen.
 
-	def __init__(self, tab):
-		QTextBrowser.__init__(self)
-		self.tab = tab
-		# if set to True, links to other files will unsuccessfully be opened as anchors
-		self.setOpenLinks(False)
-		self.anchorClicked.connect(self.openInternal)
-
-	def disconnectExternalSignals(self):
-		pass
-
-	def openInternal(self, link):
-		url = link.url()
-		isLocalHtml = (link.scheme() in ('file', '') and url.endswith('.html'))
-		if url.startswith('#'):
-			self.scrollToAnchor(url[1:])
-		elif link.isRelative() and get_markup_for_file_name(url, return_class=True):
-			fileToOpen = QDir.current().filePath(url)
-			if not QFileInfo(fileToOpen).completeSuffix() and self._fileName:
-				fileToOpen += '.' + QFileInfo(self.tab.fileName).completeSuffix()
-			self.tab.p.openFileWrapper(fileToOpen)
-		elif globalSettings.handleWebLinks and isLocalHtml:
-			self.setSource(link)
-		else:
-			QDesktopServices.openUrl(link)
+		When links like [test](test) are clicked, the file test.md is opened.
+		It has to be located next to the current opened file.
+		Relative paths like [test](../test) or [test](folder/test) are also possible.
+		"""
+		if self.fileName:
+			currentExt = splitext(self.fileName)[1]
+			basename, ext = splitext(fileToOpen)
+			if ext in ('.html', '') and exists(basename + currentExt):
+				self.p.openFileWrapper(basename + currentExt)
+				return basename + currentExt
+		if exists(fileToOpen) and get_markup_for_file_name(fileToOpen, return_class=True):
+			self.p.openFileWrapper(fileToOpen)
+			return fileToOpen
